use core::fmt;
use r2d2::ManageConnection;
use std::io::{Read, Write};
use std::{
    io::Error,
    net::{SocketAddr, TcpStream},
};

pub struct PixelflutConnectionManager {
    addr: SocketAddr,
}

pub struct PixelflutConnection {
    stream: TcpStream,
}

impl PixelflutConnectionManager {
    pub fn new(addr: &str) -> Result<Self, Error> {
        let addr: SocketAddr = match addr.parse() {
            Ok(addr) => addr,
            Err(e) => panic!("Error in parsing the socket address: {}", e),
        };
        Ok(Self { addr })
    }

    fn get_connection(&self) -> Result<PixelflutConnection, Error> {
        let stream = TcpStream::connect(self.addr)?;
        stream.set_nodelay(true)?;
        Ok(PixelflutConnection { stream })
    }
}

#[derive(Copy, Clone)]
pub struct Rgba(u8, u8, u8, u8);

impl fmt::Display for Rgba {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "{:02X}{:02X}{:02X}{:02X}",
            self.0, self.1, self.2, self.3
        )
    }
}

impl From<u32> for Rgba {
    fn from(a: u32) -> Self {
        Self((a >> 24) as u8, (a >> 16) as u8, (a >> 8) as u8, a as u8)
    }
}

impl From<image::Rgba<u8>> for Rgba {
    fn from(c: image::Rgba<u8>) -> Self {
        Self(c.0[0], c.0[1], c.0[2], c.0[3])
    }
}

impl PixelflutConnection {
    pub fn set(&mut self, x: u32, y: u32, c: impl Into<Rgba>) -> Result<(), Error> {
        let msg = format!("PX {} {} {}\n", x, y, c.into());
        self.write_command(&msg)
    }
    pub fn set_multiple(&mut self, ps: Vec<(u32, u32, impl Into<Rgba>)>) -> Result<(), Error> {
        let mut msg = String::from("");
        for (x,y,p) in ps {
            msg.push_str(&format!("PX {} {} {}\n", x, y, p.into()))
        }
        self.write_command(&msg)
    }

    pub fn get(&mut self, x: u32, y: u32) -> Result<Rgba, Error> {
        let msg = format!("PX {} {}\n", x, y);
        let mut out = String::new();

        self.write_command(&msg)?;
        self.stream.read_to_string(&mut out);
        todo!()
    }

    pub fn size(&mut self) -> Result<(usize, usize), Error> {
        let mut out = String::new();

        self.write_command("SIZE\n")?;
        self.stream.read_to_string(&mut out);
        todo!()
    }

    fn write_command(&mut self, cmd: &str) -> Result<(), Error> {
        let _ = self.stream.write(&cmd[..].as_bytes())?;
        Ok(())
    }
}

impl ManageConnection for PixelflutConnectionManager {
    type Connection = PixelflutConnection;
    type Error = Error;

    fn connect(&self) -> Result<Self::Connection, Self::Error> {
        self.get_connection()
    }

    fn is_valid(&self, conn: &mut Self::Connection) -> Result<(), Self::Error> {
        Ok(())
    }

    fn has_broken(&self, _: &mut Self::Connection) -> bool {
        false
    }
}
